'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var _require = require('path'),
    joinPath = _require.join;

var _require2 = require('fs-extra'),
    readJsonSync = _require2.readJsonSync,
    readdirSync = _require2.readdirSync,
    existsSync = _require2.existsSync;

var _require3 = require('child-process-promise'),
    spawn = _require3.spawn;

// Delve into the FS and find the names of dependencies. This includes
// uncoped names where we can just grab the name of the directory inside
// `node_modules` and scoped packages where we need to delve further
// into the scope directory, e.g. `node_modules/@versal/`.


var findDependencyNamesForLocalPackage = function findDependencyNamesForLocalPackage(rootPath) {
  var scopedPackagesRoot = joinPath(rootPath, 'node_modules');
  // Skip if we don't have any node modules present
  if (!existsSync(scopedPackagesRoot)) {
    return [];
  }

  // Find the packages directly in node_modules
  var unscopedPackages = readdirSync(scopedPackagesRoot).filter(function (moduleName) {
    return !moduleName.startsWith('@');
  }).map(function (moduleName) {
    return joinPath(rootPath, moduleName);
  });

  // Find packages in named scopes, e.g. @versal
  var scopedPackages = readdirSync(scopedPackagesRoot).reduce(function (names, moduleName) {
    if (moduleName.startsWith('@')) {
      var scopeName = moduleName.split('/').pop();
      var scopePath = joinPath(scopedPackagesRoot, moduleName);
      var scopedNames = readdirSync(scopePath).map(function (name) {
        return scopeName + '/' + name;
      });
      names = [].concat(_toConsumableArray(names), _toConsumableArray(scopedNames));
    }
    return names;
  }, []);

  return [].concat(_toConsumableArray(scopedPackages), _toConsumableArray(unscopedPackages));
};

// Finds all modules in `process.cwd()` and dredges up the manifest and other
// pertinent metadata.
var findLocalModules = function findLocalModules(rootPath) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var modulesByName = {};
  var singlePackage = config.singlePackage;


  readdirSync(rootPath)
  // Resolve the module paths
  .map(function (modulePath) {
    return {
      path: joinPath(rootPath, modulePath),
      manifestPath: joinPath(rootPath, modulePath, 'package.json')
    };
  })
  // Filter out non-modules
  .filter(function (module) {
    return existsSync(module.manifestPath);
  })
  // Get the manifest
  .map(function (module) {
    return _extends({}, module, { manifest: readJsonSync(module.manifestPath) });
  })
  // In focus mode mark skippable modules
  .map(function (module) {
    return _extends({}, module, {
      isSkippable: singlePackage && singlePackage !== module.manifest.name
    });
  })
  // Filter out un-named modules
  .filter(function (module) {
    return module.manifest.name;
  })
  // Find applicable links and add them to the registry
  .forEach(function (module) {
    var dependencyNames = findDependencyNamesForLocalPackage(module.path);
    if (modulesByName[module.manifest.name]) {
      console.warn('WARNING: Module already found');
      console.warn('  Using module found at:', modulesByName[module.manifest.name].path);
      console.warn('  Ignoring module found at:', module.path);
    } else {
      modulesByName[module.manifest.name] = _extends({}, module, { dependencyNames: dependencyNames });
    }
  });

  return modulesByName;
};

// Checks each provided module for sub-dependencies that are also found in
// provided modules. Returns an object useful for creating yarn links
// between packages. e.g.
var findLocalDependenciesForModules = function findLocalDependenciesForModules(modulesByName) {
  var moduleNames = Object.keys(modulesByName);
  var modules = moduleNames.map(function (name) {
    return modulesByName[name];
  });

  var dependencies = [];
  modules.forEach(function (_ref) {
    var name = _ref.manifest.name,
        path = _ref.path,
        isSkippable = _ref.isSkippable,
        dependencyNames = _ref.dependencyNames;

    var localDeps = dependencyNames.reduce(function (paths, depName) {
      if (moduleNames.includes(depName)) {
        var depPath = modulesByName[depName].path;

        paths.push({ path: depPath, name: depName });
      }
      return paths;
    }, []);

    if (localDeps.length) {
      dependencies.push({ dependencies: localDeps, path: path, name: name, isSkippable: isSkippable });
    }
  });

  return dependencies;
};

// Helpers for building up command strings and arguments

var buildUnlinkCommandsForLocalDependencies = function buildUnlinkCommandsForLocalDependencies(localDependencies) {
  return localDependencies.reduce(function (commands, localDependency) {
    if (localDependency.isSkippable) return commands;

    localDependency.dependencies.forEach(function (dependency) {
      commands.push(['yarn unlink --force', { cwd: dependency.path }]);
    });
    return commands;
  }, []);
};

var buildLinkCommandsForLocalDependencies = function buildLinkCommandsForLocalDependencies(localDependencies) {
  return localDependencies.reduce(function (commands, localDependency) {
    if (localDependency.isSkippable) return commands;

    localDependency.dependencies.forEach(function (dependency) {
      commands.push(['yarn link', { cwd: dependency.path }]);
    });
    return commands;
  }, []);
};

var buildLinkCommandsForLocalDependees = function buildLinkCommandsForLocalDependees(localDependencies) {
  return localDependencies.reduce(function (commands, localDependency) {
    if (localDependency.isSkippable) return commands;

    localDependency.dependencies.forEach(function (dependency) {
      commands.push(['yarn link ' + dependency.name, { cwd: localDependency.path }]);
    });
    return commands;
  }, []);
};

// Turn command strings and arguments into promise returning functions that
// will invoke the command

var buildJobsForLinkCommands = function buildJobsForLinkCommands(linkCommands) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var handleJobError = function handleJobError(error) {
    if (!config.failSilently) throw error;
  };
  return linkCommands.reduce(function (jobs, _ref2) {
    var _ref3 = _slicedToArray(_ref2, 2),
        commandString = _ref3[0],
        options = _ref3[1];

    var commandParts = commandString.split(' ');
    var command = commandParts[0];
    var args = commandParts.slice(1);
    var job = function job() {
      return spawn(command, args, options).catch(handleJobError);
    };
    return [].concat(_toConsumableArray(jobs), [job]);
  }, []);
};

// Flow control utils

var series = function series(jobs) {
  return jobs.reduce(function (runningJobs, job) {
    return runningJobs.then(job);
  }, Promise.resolve());
};

var parallel = function parallel(jobs) {
  return function () {
    return Promise.all(jobs.map(function (j) {
      return j();
    }));
  };
};

var pipeOutputs = function pipeOutputs(jobs) {
  return jobs.map(function (job) {
    return function () {
      var running = job();
      running.childProcess.stdout.pipe(process.stdout);
      running.childProcess.stderr.pipe(process.stderr);
      return running;
    };
  });
};

// Display utils

var showLinkResults = function showLinkResults(localDependencies) {
  console.info('Links created');
  localDependencies.forEach(function (localDependency) {
    if (localDependency.isSkippable) return;
    console.info('* ' + localDependency.name + ' (' + localDependency.path + ')');
    localDependency.dependencies.forEach(function (dependency) {
      console.info('  - ' + dependency.name + ' (' + dependency.path + ')');
    });
  });
  console.info('Done!');
};

var showErrorAndCrash = function showErrorAndCrash(error) {
  console.error(error.message);
  console.error('Error linking local packages! Bailing!');
  process.exit(1);
};

// App

var main = function main(config) {
  var packagesRoot = config.packagesRoot,
      verbose = config.verbose;

  var localModules = findLocalModules(packagesRoot, config);
  var localDependencies = findLocalDependenciesForModules(localModules);

  var unlinkDependencyCommands = buildUnlinkCommandsForLocalDependencies(localDependencies, config);
  var linkDependencyCommands = buildLinkCommandsForLocalDependencies(localDependencies, config);
  var linkDependeeCommands = buildLinkCommandsForLocalDependees(localDependencies, config);

  var unlinkJobs = buildJobsForLinkCommands(unlinkDependencyCommands, Object.assign({ failSilently: true }, config));

  var linkDependencyJobs = buildJobsForLinkCommands(linkDependencyCommands, Object.assign({ failSilently: true }, config));

  var linkDependeeJobs = buildJobsForLinkCommands(linkDependeeCommands, config);

  series([function () {
    return console.info('Unlinking dependencies...');
  }, verbose ? parallel(pipeOutputs(unlinkJobs)) : parallel(unlinkJobs), function () {
    return console.info('Linking dependencies...');
  }, verbose ? parallel(pipeOutputs(linkDependencyJobs)) : parallel(linkDependencyJobs), function () {
    return console.info('Linking dependees...');
  }, verbose ? parallel(pipeOutputs(linkDependeeJobs)) : parallel(linkDependeeJobs)]).then(function () {
    return showLinkResults(localDependencies);
  }).catch(showErrorAndCrash);
};

// CLI

var configureApp = require('yargs');

var config = configureApp.version().usage('Usage: link-packages [options]').option('packages-root', {
  type: 'string',
  desc: 'Path to packages',
  default: process.cwd()
}).option('single-package', {
  type: 'string',
  desc: 'Link a single dependee by path'
}).option('verbose', {
  type: 'boolean',
  desc: 'Show yarn output',
  default: false
}).help().argv;

main(config);